@@ -1,6 +1,5 @@
 /*
-   Copyright (c) 2003-2007 MySQL AB
-   Use is subject to license terms.
+   Copyright (c) 2003, 2016, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -21,10 +20,21 @@
 
 #include <NdbOut.hpp>
 #include <mgmapi_config_parameters.h>
+#include <ndb_version.h>
+
+#define JAM_FILE_ID 0
+
 
 class NodeInfo {
 public:
-  NodeInfo();
+  NodeInfo()
+  : m_version(0),
+    m_mysql_version(0),
+    m_lqh_workers(0),
+    m_type(INVALID),
+    m_connectCount(0),
+    m_connected(FALSE)
+  {}
 
   /**
    * NodeType
@@ -36,33 +46,45 @@ public:
     INVALID = 255 ///< Invalid type
   };
   NodeType getType() const;
-  
+
   Uint32 m_version;       ///< Ndb version
   Uint32 m_mysql_version; ///< MySQL version
+  Uint32 m_lqh_workers;   ///< LQH workers
   Uint32 m_type;          ///< Node type
   Uint32 m_connectCount;  ///< No of times connected
-  bool   m_connected;     ///< Node is connected
-  Uint32 m_heartbeat_cnt; ///< Missed heartbeats
-  
+  Uint32 m_connected;     ///< Node is connected
+
   friend NdbOut & operator<<(NdbOut&, const NodeInfo&); 
 };
 
 
-inline
-NodeInfo::NodeInfo(){
-  m_version = 0;
-  m_mysql_version = 0;
-  m_type = INVALID;
-  m_connectCount = 0;
-  m_heartbeat_cnt= 0;
-}
-
 inline
 NodeInfo::NodeType
 NodeInfo::getType() const {
   return (NodeType)m_type;
 }
 
+
+class NdbVersion {
+  Uint32 m_ver;
+public:
+  NdbVersion(Uint32 ver) : m_ver(ver) {};
+
+  friend NdbOut& operator<<(NdbOut&, const NdbVersion&);
+};
+
+
+inline
+NdbOut&
+operator<<(NdbOut& ndbout, const NdbVersion& ver){
+  ndbout.print("%d.%d.%d",
+               ndbGetMajor(ver.m_ver),
+               ndbGetMinor(ver.m_ver),
+               ndbGetBuild(ver.m_ver));
+  return ndbout;
+}
+
+
 inline
 NdbOut &
 operator<<(NdbOut& ndbout, const NodeInfo & info){
@@ -85,8 +107,8 @@ operator<<(NdbOut& ndbout, const NodeInfo & info){
     break;
   }
 
-  ndbout << " ndb version: " << info.m_version
-	 << " mysql version; " << info.m_mysql_version
+  ndbout << " ndb version: " << NdbVersion(info.m_version)
+	 << " mysql version: " << NdbVersion(info.m_mysql_version)
 	 << " connect count: " << info.m_connectCount
 	 << "]";
   return ndbout;
@@ -102,4 +124,7 @@ struct NodeVersionInfo
   } m_type [3]; // Indexed as NodeInfo::Type 
 };
 
+
+#undef JAM_FILE_ID
+
 #endif
