@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2014, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -20,6 +20,9 @@
 
 #include "SignalData.hpp"
 
+#define JAM_FILE_ID 137
+
+
 /**
  * DumpStateOrd is sent by the mgmtsrvr to CMVMI.
  * CMVMI the redirect the signal to all blocks.
@@ -95,6 +98,7 @@ public:
     CmvmiMaintLockCPU = 505,
     CmvmiSchedulerSpinTimer = 506,
     // 1222-1225 DICT
+    DictDumpLockQueue = 1228,
     LqhDumpAllDefinedTabs = 1332,
     LqhDumpNoLogPages = 1333,
     LqhDumpOneScanRec = 2300,
@@ -123,6 +127,7 @@ public:
     TcSetApplTransactionTimeout = 2508,
     TcStartDumpIndexOpCount = 2512,
     TcDumpIndexOpCount = 2513,
+    TcDumpApiConnectRecSummary = 2514,
     TcDumpPoolLevels = 2555,
     CmvmiDumpConnections = 2600,
     CmvmiDumpLongSignalMemory = 2601,
@@ -134,6 +139,7 @@ public:
                                       in clusterlog */
     CmvmiTestLongSig = 2605,  /* Long signal testing trigger */
     DumpEventLog = 2606,
+    CmvmiSetKillerWatchdog = 2610,
     LCPContinue = 5900,
     // 7000 DIH
     // 7001 DIH
@@ -155,12 +161,25 @@ public:
     DihAllAllowNodeStart = 7016,
     DihMinTimeBetweenLCP = 7017,
     DihMaxTimeBetweenLCP = 7018,
-    // Checks whether add frag failure was cleaned up.
-    // Should NOT be used while commands involving addFragReq
-    // are being performed
-    DihAddFragFailCleanedUp = 7019,
+    // Check if blocks are done with handling the failure of another node.
+    DihTcSumaNodeFailCompleted = 7019, // DIH+TC+SUMA
     // 7020
     // 7021
+    // 7022
+    // 7023
+    /*
+      Checks whether add frag failure was cleaned up.
+      Should NOT be used while commands involving addFragReq
+      are being performed.
+      NB: This value is only intended for use in test cases. If used 
+      interactively, it is likely to crash the node. It should therefore
+      *not* be described in end-user documentation.
+    */
+    DihAddFragFailCleanedUp = 7024,
+    /**
+     * Allows GCP stop thresholds to be set
+     */
+    DihSetGcpStopVals = 7026,
     DihDumpPageRecInfo = 7032,
     EnableUndoDelayDataWrite = 7080, // DIH+ACC+TUP
     DihSetTimeBetweenGcp = 7090,
@@ -195,4 +214,7 @@ public:
   Uint32 args[25];          // Generic argument
 };
 
+
+#undef JAM_FILE_ID
+
 #endif
