@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -20,6 +20,10 @@
 
 #include <ndb_global.h>
 
+#if defined(HAVE__BITSCANFORWARD) || defined(HAVE__BITSCANREVERSE)
+#include <intrin.h>
+#endif
+
 /**
  * Bitmask implementation.  Size (in 32-bit words) is given explicitly
  * (as first argument).  All methods are static.
@@ -30,9 +34,15 @@ public:
 
   /**
    * get - Check if bit n is set.
+   *       assert(n < (32 * size))
    */
   static bool get(unsigned size, const Uint32 data[], unsigned n);
 
+  /**
+   * safe_get - Check if bit N is set, accept any value for N
+   */
+  static bool safe_get(unsigned size, const Uint32 data[], unsigned n);
+
   /**
    * set - Set bit n to given value (true/false).
    */
@@ -88,6 +98,54 @@ public:
    */
   static unsigned count(unsigned size, const Uint32 data[]);
 
+  /**
+   * return count trailing zero bits inside a word
+   * undefined behaviour if non set
+   */
+  static unsigned ctz(Uint32 x);
+
+  /**
+   * return count leading zero bits inside a word
+   * undefined behaviour if non set
+   */
+  static unsigned clz(Uint32 x);
+
+  /**
+   * return index of first bit set inside a word
+   * undefined behaviour if non set
+   */
+  static unsigned ffs(Uint32 x);
+
+  /**
+   * return index of last bit set inside a word
+   * undefined behaviour if non set
+   */
+  static unsigned fls(Uint32 x);
+
+  /**
+   * find - Find first set bit, starting from 0
+   * Returns NotFound when not found.
+   */
+  static unsigned find_first(unsigned size, const Uint32 data[]);
+
+  /**
+   * find - Find last set bit, starting from 0
+   * Returns NotFound when not found.
+   */
+  static unsigned find_last(unsigned size, const Uint32 data[]);
+
+  /**
+   * find - Find first set bit, starting at given position.
+   * Returns NotFound when not found.
+   */
+  static unsigned find_next(unsigned size, const Uint32 data[], unsigned n);
+
+  /**
+   * find - Find last set bit, starting at given position.
+   * Returns NotFound when not found.
+   */
+  static unsigned find_prev(unsigned size, const Uint32 data[], unsigned n);
+
   /**
    * find - Find first set bit, starting at given position.
    * Returns NotFound when not found.
@@ -179,15 +237,14 @@ public:
    */
   static char* getText(unsigned size, const Uint32 data[], char* buf);
 
-  /**
-   * Parse string with numbers format
-   *   1,2,3-5
-   * @return -1 if unparsable chars found, 
-   *         -2 str has number > bitmask size
-   *            else returns number of bits set 
-   */
-  static int parseMask(unsigned size, Uint32 data[], const char * str);
+  /* Fast bit counting (16 instructions on x86_64, gcc -O3). */
+  static inline Uint32 count_bits(Uint32 x);
 
+  /**
+   * store each set bit in <em>dst</em> and return bits found
+   */
+  static Uint32 toArray(Uint8* dst, Uint32 len,
+                        unsigned size, const Uint32 data[]);
 private:
   static void getFieldImpl(const Uint32 data[], unsigned, unsigned, Uint32 []);
   static void setFieldImpl(Uint32 data[], unsigned, unsigned, const Uint32 []);
@@ -200,6 +257,16 @@ BitmaskImpl::get(unsigned size, const Uint32 data[], unsigned n)
   return (data[n >> 5] & (1 << (n & 31))) != 0;
 }
 
+inline bool
+BitmaskImpl::safe_get(unsigned size, const Uint32 data[], unsigned n)
+{
+  if (n < (size << 5))
+  {
+    return (data[n >> 5] & (1 << (n & 31))) != 0;
+  }
+  return false;
+}
+
 inline void
 BitmaskImpl::set(unsigned size, Uint32 data[], unsigned n, bool value)
 {
@@ -331,25 +398,279 @@ BitmaskImpl::count(unsigned size, const Uint32 data[])
 {
   unsigned cnt = 0;
   for (unsigned i = 0; i < size; i++) {
-    Uint32 x = data[i];
-    while (x) {
-      x &= (x - 1);
-      cnt++;
-    }
+    cnt += count_bits(data[i]);
   }
   return cnt;
 }
 
+/**
+ * return count trailing zero bits inside a word
+ * undefined behaviour if non set
+ */
+inline
+Uint32
+BitmaskImpl::ctz(Uint32 x)
+{
+  return ffs(x);
+}
+
+/**
+ * return count leading bits inside a word
+ * undefined behaviour if non set
+ */
+inline
+Uint32
+BitmaskImpl::clz(Uint32 x)
+{
+#if defined HAVE___BUILTIN_CLZ
+  return __builtin_clz(x);
+#elif defined(__GNUC__) && (defined(__x86_64__) || defined (__i386__))
+  asm("bsr %1,%0"
+      : "=r" (x)
+      : "rm" (x));
+  return 31 - x;
+#elif defined HAVE__BITSCANREVERSE
+  unsigned long r;
+  unsigned char res = _BitScanReverse(&r, (unsigned long)x);
+  assert(res > 0);
+  return 31 - (Uint32)r;
+#else
+  int b = 0;
+  if (!(x & 0xffff0000))
+  {
+    x <<= 16;
+    b += 16;
+  }
+  if (!(x & 0xff000000))
+  {
+    x <<= 8;
+    b += 8;
+  }
+  if (!(x & 0xf0000000))
+  {
+    x <<= 4;
+    b += 4;
+  }
+  if (!(x & 0xc0000000))
+  {
+    x <<= 2;
+    b += 2;
+  }
+  if (!(x & 0x80000000))
+  {
+    x <<= 1;
+    b += 1;
+  }
+  return b;
+#endif
+}
+
+/**
+ * return index of first bit set inside a word
+ * undefined behaviour if non set
+ */
+inline
+Uint32
+BitmaskImpl::ffs(Uint32 x)
+{
+#if defined HAVE___BUILTIN_CTZ
+  return __builtin_ctz(x);
+#elif defined(__GNUC__) && (defined(__x86_64__) || defined (__i386__))
+  asm("bsf %1,%0"
+      : "=r" (x)
+      : "rm" (x));
+  return x;
+#elif defined HAVE___BUILTIN_FFS
+  /**
+   * gcc defined ffs(0) == 0, and returned indexes 1-32
+   */
+  return __builtin_ffs(x) - 1;
+#elif defined HAVE__BITSCANFORWARD
+  unsigned long r;
+  unsigned char res = _BitScanForward(&r, (unsigned long)x);
+  assert(res > 0);
+  return (Uint32)r;
+#elif defined HAVE_FFS
+  return ::ffs(x) - 1;
+#else
+  int b = 0;
+  if (!(x & 0xffff))
+  {
+    x >>= 16;
+    b += 16;
+  }
+  if (!(x & 0xff))
+  {
+    x >>= 8;
+    b += 8;
+  }
+  if (!(x & 0xf))
+  {
+    x >>= 4;
+    b += 4;
+  }
+  if (!(x & 3))
+  {
+    x >>= 2;
+    b += 2;
+  }
+  if (!(x & 1))
+  {
+    x >>= 1;
+    b += 1;
+  }
+  return b;
+#endif
+}
+
+/**
+ * return index of last bit set inside a word
+ * undefined behaviour if non set
+ */
+inline
+Uint32
+BitmaskImpl::fls(Uint32 x)
+{
+#if defined(__GNUC__) && (defined(__x86_64__) || defined (__i386__))
+  asm("bsr %1,%0"
+      : "=r" (x)
+      : "rm" (x));
+  return x;
+#elif defined HAVE___BUILTIN_CLZ
+  return 31 - __builtin_clz(x);
+#elif defined HAVE__BITSCANREVERSE
+  unsigned long r;
+  unsigned char res = _BitScanReverse(&r, (unsigned long)x);
+  assert(res > 0);
+  return (Uint32)r;
+#else
+  int b = 31;
+  if (!(x & 0xffff0000))
+  {
+    x <<= 16;
+    b -= 16;
+  }
+  if (!(x & 0xff000000))
+  {
+    x <<= 8;
+    b -= 8;
+  }
+  if (!(x & 0xf0000000))
+  {
+    x <<= 4;
+    b -= 4;
+  }
+  if (!(x & 0xc0000000))
+  {
+    x <<= 2;
+    b -= 2;
+  }
+  if (!(x & 0x80000000))
+  {
+    x <<= 1;
+    b -= 1;
+  }
+  return b;
+#endif
+}
+
+inline unsigned
+BitmaskImpl::find_first(unsigned size, const Uint32 data[])
+{
+  Uint32 n = 0;
+  while (n < (size << 5))
+  {
+    Uint32 val = data[n >> 5];
+    if (val)
+    {
+      return n + ffs(val);
+    }
+    n += 32;
+  }
+ return NotFound;
+}
+
+inline unsigned
+BitmaskImpl::find_last(unsigned size, const Uint32 data[])
+{
+  if (size == 0)
+    return NotFound;
+  Uint32 n = (size << 5) - 1;
+  do
+  {
+    Uint32 val = data[n >> 5];
+    if (val)
+    {
+      return n - clz(val);
+    }
+    n -= 32;
+  } while (n != 0xffffffff);
+ return NotFound;
+}
+
+inline unsigned
+BitmaskImpl::find_next(unsigned size, const Uint32 data[], unsigned n)
+{
+  assert(n <= (size << 5));
+  if (n == (size << 5)) // allow one step utside for easier use
+    return NotFound;
+  Uint32 val = data[n >> 5];
+  Uint32 b = n & 31;
+  if (b)
+  {
+    val >>= b;
+    if (val)
+    {
+      return n + ffs(val);
+    }
+    n += 32 - b;
+  }
+
+  while (n < (size << 5))
+  {
+    val = data[n >> 5];
+    if (val)
+    {
+      return n + ffs(val);
+    }
+    n += 32;
+  }
+  return NotFound;
+}
+
+inline unsigned
+BitmaskImpl::find_prev(unsigned size, const Uint32 data[], unsigned n)
+{
+  if (n >= (Uint32) 0xffffffff /* -1 */) // allow one bit outside array for easier use
+    return NotFound;
+  assert(n < (size << 5));
+  Uint32 val = data[n >> 5];
+  Uint32 b = n & 31;
+  if (b < 31)
+  {
+    val <<= 31 - b;
+    if (val)
+    {
+      return n - clz(val);
+    }
+    n -= b + 1;
+  }
+
+  while (n != NotFound) {
+    val = data[n >> 5];
+    if (val)
+    {
+      return n - clz(val);
+    }
+    n -= 32;
+  }
+  return NotFound;
+}
+
 inline unsigned
 BitmaskImpl::find(unsigned size, const Uint32 data[], unsigned n)
 {
-  while (n < (size << 5)) {             // XXX make this smarter
-    if (get(size, data, n)) {
-      return n;
-    }
-    n++;
-  }
-  return NotFound;
+  return find_next(size, data, n);
 }
 
 inline bool
@@ -463,6 +784,42 @@ BitmaskImpl::getText(unsigned size, const Uint32 data[], char* buf)
   return org;
 }
 
+inline
+Uint32
+BitmaskImpl::count_bits(Uint32 x)
+{
+  x= x - ((x>>1) & 0x55555555);
+  x= (x & 0x33333333) + ((x>>2) & 0x33333333);
+  x= (x + (x>>4)) & 0x0f0f0f0f;
+  x= (x*0x01010101) >> 24;
+  return x;
+}
+
+inline
+Uint32
+BitmaskImpl::toArray(Uint8* dst, Uint32 len,
+                     unsigned size, const Uint32 * data)
+{
+  assert(len >= size * 32);
+  assert(32 * size <= 256); // Uint8
+  Uint8 * save = dst;
+  for (Uint32 i = 0; i<size; i++)
+  {
+    Uint32 val = * data++;
+    Uint32 bit = 0;
+    while (val)
+    {
+      if (val & (1 << bit))
+      {
+        * dst++ = 32 * i + bit;
+        val &= ~(1U << bit);
+      }
+      bit ++;
+    }
+  }
+  return (Uint32)(dst - save);
+}
+
 /**
  * Bitmasks.  The size is number of 32-bit words (Uint32).
  * Unused bits in the last word must be zero.
@@ -491,6 +848,10 @@ public:
   STATIC_CONST( NotFound = BitmaskImpl::NotFound );
   STATIC_CONST( TextLength = size * 8 );
 
+  Uint32 getSizeInWords() const { return Size;}
+
+  unsigned max_size() const { return (size * 32) - 1; }
+
   /**
    * assign - Set all bits in <em>dst</em> to corresponding in <em>src/<em>
    */
@@ -513,12 +874,22 @@ public:
    */
   void assign(unsigned sz, const Uint32 src[]);
 
+  /**
+   * start of static members
+   */
+
   /**
    * get - Check if bit n is set.
    */
   static bool get(const Uint32 data[], unsigned n);
   bool get(unsigned n) const;
 
+  /**
+   * safe_get - Check if bit N is set, accept any value for N
+   */
+  static bool safe_get(const Uint32 data[], unsigned n);
+  bool safe_get(unsigned n) const;
+
   /**
    * set - Set bit n to given value (true/false).
    */
@@ -574,6 +945,34 @@ public:
   static unsigned count(const Uint32 data[]);
   unsigned count() const;
 
+  /**
+   * find - Find first set bit, starting at 0
+   * Returns NotFound when not found.
+   */
+  static unsigned find_first(const Uint32 data[]);
+  unsigned find_first() const;
+
+  /**
+   * find - Find first set bit, starting at 0
+   * Returns NotFound when not found.
+   */
+  static unsigned find_next(const Uint32 data[], unsigned n);
+  unsigned find_next(unsigned n) const;
+
+  /**
+   * find - Find last set bit, starting at 0
+   * Returns NotFound when not found.
+   */
+  static unsigned find_last(const Uint32 data[]);
+  unsigned find_last() const;
+
+  /**
+   * find - Find previous set bit, starting at n
+   * Returns NotFound when not found.
+   */
+  static unsigned find_prev(const Uint32 data[], unsigned n);
+  unsigned find_prev(unsigned n) const;
+
   /**
    * find - Find first set bit, starting at given position.
    * Returns NotFound when not found.
@@ -641,8 +1040,8 @@ public:
   static char* getText(const Uint32 data[], char* buf);
   char* getText(char* buf) const;
 
-  static int parseMask(Uint32 data[], const char * src);
-  int parseMask(const char * src);
+  static Uint32 toArray(Uint8 * dst, Uint32 len, const Uint32 data[]);
+  Uint32 toArray(Uint8 * dst, Uint32 len) const;
 };
 
 template <unsigned size>
@@ -701,6 +1100,20 @@ BitmaskPOD<size>::get(unsigned n) const
   return BitmaskPOD<size>::get(rep.data, n);
 }
 
+template <unsigned size>
+inline bool
+BitmaskPOD<size>::safe_get(const Uint32 data[], unsigned n)
+{
+  return BitmaskImpl::safe_get(size, data, n);
+}
+
+template <unsigned size>
+inline bool
+BitmaskPOD<size>::safe_get(unsigned n) const
+{
+  return BitmaskPOD<size>::safe_get(rep.data, n);
+}
+
 template <unsigned size>
 inline void
 BitmaskPOD<size>::set(Uint32 data[], unsigned n, bool value)
@@ -828,18 +1241,74 @@ BitmaskPOD<size>::count() const
   return BitmaskPOD<size>::count(rep.data);
 }
 
+template <unsigned size>
+inline unsigned
+BitmaskPOD<size>::find_first(const Uint32 data[])
+{
+  return BitmaskImpl::find_first(size, data);
+}
+
+template <unsigned size>
+inline unsigned
+BitmaskPOD<size>::find_first() const
+{
+  return BitmaskPOD<size>::find_first(rep.data);
+}
+
+template <unsigned size>
+inline unsigned
+BitmaskPOD<size>::find_next(const Uint32 data[], unsigned n)
+{
+  return BitmaskImpl::find_next(size, data, n);
+}
+
+template <unsigned size>
+inline unsigned
+BitmaskPOD<size>::find_next(unsigned n) const
+{
+  return BitmaskPOD<size>::find_next(rep.data, n);
+}
+
+template <unsigned size>
+inline unsigned
+BitmaskPOD<size>::find_last(const Uint32 data[])
+{
+  return BitmaskImpl::find_last(size, data);
+}
+
+template <unsigned size>
+inline unsigned
+BitmaskPOD<size>::find_last() const
+{
+  return BitmaskPOD<size>::find_last(rep.data);
+}
+
+template <unsigned size>
+inline unsigned
+BitmaskPOD<size>::find_prev(const Uint32 data[], unsigned n)
+{
+  return BitmaskImpl::find_prev(size, data, n);
+}
+
+template <unsigned size>
+inline unsigned
+BitmaskPOD<size>::find_prev(unsigned n) const
+{
+  return BitmaskPOD<size>::find_prev(rep.data, n);
+}
+
 template <unsigned size>
 inline unsigned
 BitmaskPOD<size>::find(const Uint32 data[], unsigned n)
 {
-  return BitmaskImpl::find(size, data, n);
+  return find_next(data, n);
 }
 
 template <unsigned size>
 inline unsigned
 BitmaskPOD<size>::find(unsigned n) const
 {
-  return BitmaskPOD<size>::find(rep.data, n);
+  return find_next(n);
 }
 
 template <unsigned size>
@@ -989,18 +1458,19 @@ BitmaskPOD<size>::overlaps(BitmaskPOD<size> that)
 }
 
 template <unsigned size>
-int
-BitmaskPOD<size>::parseMask(Uint32 data[], const char* buf)
+inline
+Uint32
+BitmaskPOD<size>::toArray(Uint8* dst, Uint32 len, const Uint32 data[])
 {
-  return BitmaskImpl::parseMask(size, data, buf);
+  return BitmaskImpl::toArray(dst, len, size, data);
 }
 
 template <unsigned size>
 inline
-int
-BitmaskPOD<size>::parseMask(const char* buf)
+Uint32
+BitmaskPOD<size>::toArray(Uint8* dst, Uint32 len) const
 {
-  return BitmaskPOD<size>::parseMask(rep.data, buf);
+  return BitmaskImpl::toArray(dst, len, size, this->rep.data);
 }
 
 template <unsigned size>
@@ -1026,6 +1496,25 @@ public:
     }
     return * this;
   }
+
+  template<unsigned sz2> Bitmask& operator=(const BitmaskPOD<sz2>& src){
+    if (size >= sz2)
+    {
+      for (unsigned i = 0; i < sz2; i++) 
+      {
+	this->rep.data[i] = src.rep.data[i];
+      }
+    }
+    else
+    {
+      assert(src.find(32*size+1) == BitmaskImpl::NotFound);
+      for (unsigned i = 0; i < size; i++) 
+      {
+	this->rep.data[i] = src.rep.data[i];
+      }
+    }
+    return * this;
+  }
 };
 
 inline void
