@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2014, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -18,23 +18,15 @@
 
 #include <ndb_global.h>
 
+#include "API.hpp"
 #include "NdbApiSignal.hpp"
 #include "NdbImpl.hpp"
-#include <NdbOperation.hpp>
-#include <NdbTransaction.hpp>
-#include <NdbRecAttr.hpp>
-#include <IPCConfig.hpp>
-#include "TransporterFacade.hpp"
 #include <ConfigRetriever.hpp>
 #include <ndb_limits.h>
 #include <NdbOut.hpp>
 #include <NdbSleep.h>
 #include "ObjectMap.hpp"
-#include <NdbIndexScanOperation.hpp>
-#include <NdbIndexOperation.hpp>
 #include "NdbUtil.hpp"
-#include <NdbBlob.hpp>
-#include "NdbEventOperationImpl.hpp"
 
 #include <EventLogger.hpp>
 extern EventLogger * g_eventLogger;
@@ -69,7 +61,8 @@ void Ndb::setup(Ndb_cluster_connection *ndb_cluster_connection,
   theMinNoOfEventsToWakeUp= 0;
   theTransactionList= NULL;
   theConnectionArray= NULL;
-  the_last_check_time= 0;
+  theConnectionArrayLast= NULL;
+  the_last_check_time = NdbTick_CurrentMillisecond();
   theFirstTransId= 0;
   theRestartGCI= 0;
   theNdbBlockNumber= -1;
@@ -79,9 +72,6 @@ void Ndb::setup(Ndb_cluster_connection *ndb_cluster_connection,
   theFirstTransId= 0;
   theMyRef= 0;
 
-  cond_wait_index = TransporterFacade::MAX_NO_THREADS;
-  cond_signal_ndb = NULL;
-
   fullyQualifiedNames = true;
 
 #ifdef POORMANSPURIFY
@@ -94,20 +84,21 @@ void Ndb::setup(Ndb_cluster_connection *ndb_cluster_connection,
   theError.code = 0;
 
   theConnectionArray = new NdbConnection * [MAX_NDB_NODES];
+  theConnectionArrayLast = new NdbConnection * [MAX_NDB_NODES];
   theCommitAckSignal = NULL;
+  theCachedMinDbNodeVersion = 0;
   
   int i;
   for (i = 0; i < MAX_NDB_NODES ; i++) {
     theConnectionArray[i] = NULL;
-  }//forg
+    theConnectionArrayLast[i] = NULL;
+  }//for
   m_sys_tab_0 = NULL;
 
   theImpl->m_dbname.assign(aDataBase);
   theImpl->m_schemaname.assign(aSchema);
   theImpl->update_prefix();
 
-  theImpl->theWaiter.m_mutex =  theImpl->m_transporter_facade->theMutexPtr;
-
   // Signal that the constructor has finished OK
   if (theInitState == NotConstructed)
     theInitState = NotInitialised;
@@ -137,6 +128,14 @@ Ndb::~Ndb()
   DBUG_ENTER("Ndb::~Ndb()");
   DBUG_PRINT("enter",("this: 0x%lx", (long) this));
 
+  if (theImpl == NULL)
+  {
+    /* Help users find their bugs */
+    g_eventLogger->warning("Deleting Ndb-object @%p which is already deleted?",
+                           this);
+    DBUG_VOID_RETURN;
+  }
+
   if (m_sys_tab_0)
     getDictionary()->removeTableGlobal(*m_sys_tab_0, 0);
 
@@ -163,15 +162,17 @@ Ndb::~Ndb()
   doDisconnect();
 
   /* Disconnect from transporter to stop signals from coming in */
-  if (theImpl->m_transporter_facade != NULL && theNdbBlockNumber > 0){
-    theImpl->m_transporter_facade->close(theNdbBlockNumber, theFirstTransId);
-  }
+  theImpl->close();
 
   delete theEventBuffer;
+  theEventBuffer = NULL;
 
   releaseTransactionArrays();
 
   delete []theConnectionArray;
+  theConnectionArray = NULL;
+  delete []theConnectionArrayLast;
+  theConnectionArrayLast = NULL;
   if(theCommitAckSignal != NULL){
     delete theCommitAckSignal; 
     theCommitAckSignal = NULL;
@@ -180,6 +181,7 @@ Ndb::~Ndb()
   theImpl->m_ndb_cluster_connection.unlink_ndb_object(this);
 
   delete theImpl;
+  theImpl = NULL;
 
 #ifdef POORMANSPURIFY
 #ifdef POORMANSGUI
@@ -195,17 +197,16 @@ Ndb::~Ndb()
   DBUG_VOID_RETURN;
 }
 
-NdbWaiter::NdbWaiter(){
+NdbWaiter::NdbWaiter(trp_client* clnt)
+  : m_clnt(clnt)
+{
   m_node = 0;
   m_state = NO_WAIT;
-  m_mutex = 0;
-  m_poll_owner= false;
-  m_cond_wait_index= TransporterFacade::MAX_NO_THREADS;
-  m_condition = NdbCondition_Create();
 }
 
-NdbWaiter::~NdbWaiter(){
-  NdbCondition_Destroy(m_condition);
+NdbWaiter::~NdbWaiter()
+{
+  m_clnt = NULL;
 }
 
 NdbImpl::NdbImpl(Ndb_cluster_connection *ndb_cluster_connection,
@@ -217,11 +218,13 @@ NdbImpl::NdbImpl(Ndb_cluster_connection *ndb_cluster_connection,
     m_transporter_facade(ndb_cluster_connection->m_impl.m_transporter_facade),
     m_dictionary(ndb),
     theCurrentConnectIndex(0),
-    theNdbObjectIdMap(m_transporter_facade->theMutexPtr,
-		      1024,1024),
+    theNdbObjectIdMap(1024,1024),
     theNoOfDBnodes(0),
+    theWaiter(this),
+    wakeHandler(0),
+    wakeContext(~Uint32(0)),
     m_ev_op(0),
-    customDataPtr(0)
+    customData(0)
 {
   int i;
   for (i = 0; i < MAX_NDB_NODES; i++) {
@@ -232,9 +235,22 @@ NdbImpl::NdbImpl(Ndb_cluster_connection *ndb_cluster_connection,
 
   m_systemPrefix.assfmt("%s%c%s%c", NDB_SYSTEM_DATABASE, table_name_separator,
 			NDB_SYSTEM_SCHEMA, table_name_separator);
+
+  forceShortRequests = false;
+  const char* f= getenv("NDB_FORCE_SHORT_REQUESTS");
+  if (f != 0 && *f != 0 && *f != '0' && *f != 'n' && *f != 'N')
+    forceShortRequests = true;
+
+  for (i = 0; i < Ndb::NumClientStatistics; i++)
+    clientStats[i] = 0;
 }
 
 NdbImpl::~NdbImpl()
 {
+  m_next_ndb_object = NULL;
+  m_prev_ndb_object = NULL;
+  theWaiter = NULL;
+  wakeHandler = NULL;
+  m_ev_op = NULL;
 }
 
