@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2004, 2011, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2004, 2014, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -98,6 +98,10 @@ static const char* empty_string = "";
  * 4700 - "" Event
  * 4800 - API, QueryBuilder
  * 5000 - Management server
+ * 6000 - 6999 User error codes, to be used with 
+ *   NdbInterpretedCode::interpret_exit_nok(). Do not define internal error 
+ *   codes in this range!
+ * 20000 - SPJ
  */
 
 static
@@ -129,19 +133,48 @@ ErrorBundle ErrorCodes[] = {
   {  499, DMEC, NR, "Scan take over error, restart scan transaction" },  
   { 1204, DMEC, NR, "Temporary failure, distribution changed" },
   { 4002, DMEC, NR, "Send to NDB failed" },
+  { 4007, DMEC, NR, "Send to ndbd node failed" },
   { 4010, DMEC, NR, "Node failure caused abort of transaction" }, 
+  { 4013, DMEC, NR, "Request timed out in waiting for node failure"}, 
   { 4025, DMEC, NR, "Node failure caused abort of transaction" }, 
   { 4027, DMEC, NR, "Node failure caused abort of transaction" },
   { 4028, DMEC, NR, "Node failure caused abort of transaction" },
   { 4029, DMEC, NR, "Node failure caused abort of transaction" },
   { 4031, DMEC, NR, "Node failure caused abort of transaction" },
   { 4033, DMEC, NR, "Send to NDB failed" },
+  { 4035, DMEC, NR, "Cluster temporary unavailable" },
   { 4115, DMEC, NR, 
     "Transaction was committed but all read information was not "
     "received due to node crash" },
   { 4119, DMEC, NR, "Simple/dirty read failed due to node failure" },
+
+  /**
+   * SPJ error codes
+   */ 
+
+  { 20000, DMEC, TR, "Query aborted due out of operation records" },
+  { 20001, DMEC, IE, "Query aborted due to empty query tree" },
+  { 20002, DMEC, IE, "Query aborted due to invalid request" },
+  { 20003, DMEC, IE, "Query aborted due to  unknown query operation" },
+  { 20004, DMEC, IE, "Query aborted due to invalid tree node specification" },
+  { 20005, DMEC, IE, "Query aborted due to invalid tree parameter specification" },
+  { 20006, DMEC, TR, "Query aborted due to out of LongMessageBuffer" },
+  { 20007, DMEC, IE, "Query aborted due to invalid pattern" },
+  { 20008, DMEC, TR, "Query aborted due to out of query memory" },
+  { 20009, DMEC, IE, "Query aborted due to query node too big" },
+  { 20010, DMEC, IE, "Query aborted due to query node parameters too big" },
+  { 20011, DMEC, IE, "Query aborted due to both tree and parameters contain interpreted program" },
+  { 20012, DMEC, IE, "Query aborted due to invalid tree parameter specification: Key parameter bits mismatch" },
+  { 20013, DMEC, IE, "Query aborted due to invalid tree parameter specification: Incorrect key parameter count" },
+  { 20014, DMEC, IE, "Query aborted due to internal error" },
+  { 20015, DMEC, TR, "Query aborted due to out of row memory" },
   { 20016, DMEC, NR, "Query aborted due to node failure" },
-  
+  { 20017, DMEC, IE, "Query aborted due to invalid node count" },
+  { 20018, DMEC, IE, "Query aborted due to index fragment not found" },
+  { 20019, HA_ERR_NO_SUCH_TABLE, SE, "Query table not defined" },
+  { 20020, HA_ERR_NO_SUCH_TABLE, SE, "Query table is being dropped" },
+  { 20021, HA_ERR_TABLE_DEF_CHANGED, SE, "Query table definition has changed" },
+
   /**
    * Node shutdown
    */
@@ -157,15 +190,31 @@ ErrorBundle ErrorCodes[] = {
   
   /**
    * Unknown result
+   * 
+   * We want to avoid reporting these error codes as much as possible. There
+   * are two cases where we report this as the error code.
+   *
+   * 1) We have sent a request to NDB, but for some reason we got no response,
+   *    the node is still alive and the send was successful, so what happened
+   *    is simply unknown, it shouldn't happen, most likely it is caused by
+   *    some bug somewhere.
+   *    4008 and 4012 are indications of this problem.
+   *
+   * 2) We have no connection to the cluster at all or all nodes we're
+   *    connected to are shutting down. So we have no communication to
+   *    the cluster. We will avoid reporting this error if we even only
+   *    have a starting node that we're connected since this is and
+   *    indication that we're very close to having a cluster up and
+   *    running again.
+   *    The cluster can still be up, but our API node have no ability to
+   *    see any nodes being up, we don't know whether this depends on the
+   *    cluster actually being down or if we simply have no communication
+   *    link to it at present.
    */
-  { 4007, DMEC, UR, "Send to ndbd node failed" },
   { 4008, DMEC, UR, "Receive from NDB failed" },
   { 4009, HA_ERR_NO_CONNECTION, UR, "Cluster Failure" },
   { 4012, DMEC, UR, 
     "Request ndbd time-out, maybe due to high load or communication problems"}, 
-  { 4013, DMEC, UR, "Request timed out in waiting for node failure"}, 
-  { 4024, DMEC, UR, 
-    "Time-out, most likely caused by simple read or cluster failure" }, 
   
   /**
    * TemporaryResourceError
@@ -173,10 +222,14 @@ ErrorBundle ErrorCodes[] = {
   { 217,  DMEC, TR, "217" },
   { 218,  DMEC, TR, "Out of LongMessageBuffer" },
   { 219,  DMEC, TR, "219" },
+  { 221,  DMEC, TR, "Too many concurrently fired triggers (increase "
+    "MaxNoOfFiredTriggers)" },
   { 233,  DMEC, TR,
     "Out of operation records in transaction coordinator (increase MaxNoOfConcurrentOperations)" },
   { 275,  DMEC, TR, "Out of transaction records for complete phase (increase MaxNoOfConcurrentTransactions)" },
   { 279,  DMEC, TR, "Out of transaction markers in transaction coordinator" },
+  { 273,  DMEC, TR, "Out of transaction markers databuffer in transaction coordinator" },
+  { 312,  DMEC, TR, "Out of LongMessageBuffer" },
   { 414,  DMEC, TR, "414" },
   { 418,  DMEC, TR, "Out of transaction buffers in LQH" },
   { 419,  DMEC, TR, "419" },
@@ -330,7 +383,7 @@ ErrorBundle ErrorCodes[] = {
   { 831,  DMEC, AE, "Too many nullable/bitfields in table definition" },
   { 850,  DMEC, AE, "Too long or too short default value"},
   { 851,  DMEC, AE, "Maximum 8052 bytes of FIXED columns supported"
-    ", use varchar or COLUMN_FORMAT DYNMIC instead" },
+    ", use varchar or COLUMN_FORMAT DYNAMIC instead" },
   { 876,  DMEC, AE, "876" },
   { 877,  DMEC, AE, "877" },
   { 878,  DMEC, AE, "878" },
@@ -407,6 +460,7 @@ ErrorBundle ErrorCodes[] = {
   { 708,  DMEC, SE, "No more attribute metadata records (increase MaxNoOfAttributes)" },
   { 709,  HA_ERR_NO_SUCH_TABLE, SE, "No such table existed" },
   { 710,  DMEC, SE, "Internal: Get by table name not supported, use table id." },
+  { 712,  DMEC, SE, "No more hashmap metadata records" },
   { 721,  HA_ERR_TABLE_EXIST,   OE, "Schema object with given name already exists" },
   { 723,  HA_ERR_NO_SUCH_TABLE, SE, "No such table existed" },
   { 736,  DMEC, SE, "Unsupported array size" },
@@ -458,7 +512,7 @@ ErrorBundle ErrorCodes[] = {
   { 908,  DMEC, IS, "Invalid ordered index tree node size" },
   { 909,  DMEC, IE, "No free index scan op" },
   { 910, HA_ERR_NO_SUCH_TABLE, SE, "Index is being dropped" },
-  { 913,  DMEC, AE, "Invalid index for index state update" },
+  { 913,  DMEC, AE, "Invalid index for index stats update" },
   { 914,  DMEC, IE, "Invalid index stats request" },
   { 915,  DMEC, TR, "No free index stats op" },
   { 916,  DMEC, IE, "Invalid index stats sys tables" },
@@ -467,7 +521,7 @@ ErrorBundle ErrorCodes[] = {
   { 919,  DMEC, TR, "Cannot execute index stats update" },
   { 1224, HA_WRONG_CREATE_OPTION, SE, "Too many fragments" },
   { 1225, DMEC, SE, "Table not defined in local query handler" },
-  { 1226, DMEC, SE, "Table is being dropped" },
+  { 1226, HA_ERR_NO_SUCH_TABLE, SE, "Table is being dropped" },
   { 1227, HA_WRONG_CREATE_OPTION, SE, "Invalid schema version" },
   { 1228, DMEC, SE, "Cannot use drop table for drop index" },
   { 1229, DMEC, SE, "Too long frm data supplied" },
@@ -522,7 +576,7 @@ ErrorBundle ErrorCodes[] = {
   { 1323, DMEC, IE, "1323" },
   { 1324, DMEC, IE, "Backup log buffer full" },
   { 1325, DMEC, IE, "File or scan error" },
-  { 1326, DMEC, IE, "Backup abortet due to node failure" },
+  { 1326, DMEC, IE, "Backup aborted due to node failure" },
   { 1327, DMEC, IE, "1327" },
   
   { 1340, DMEC, IE, "Backup undefined error" },
@@ -588,7 +642,7 @@ ErrorBundle ErrorCodes[] = {
   { 1426, DMEC, SE, "No such subscriber" },
   { 1427, DMEC, NR, "Api node died, when SUB_START_REQ reached node "},
   { 1428, DMEC, IE, "No replica to scan on this node (internal index stats error)" },
-  { 1429, DMEC, IE, "Subscriber node undefined in SubStartReq (config change??)" },
+  { 1429, DMEC, IE, "Subscriber node undefined in SubStartReq (config change?)" },
 
   { 4004, DMEC, AE, "Attribute name or id not found in the table" },
   
@@ -597,6 +651,8 @@ ErrorBundle ErrorCodes[] = {
   { 4102, DMEC, AE, "Type in NdbTamper not correct" },
   { 4103, DMEC, AE, "No schema connections to NDB available and connect failed" },
   { 4104, DMEC, AE, "Ndb Init in wrong state, destroy Ndb object and create a new" },
+  { 4121, DMEC, AE, "Cannot set name twice for an Ndb object" },
+  { 4122, DMEC, AE, "Cannot set name after Ndb object is initialised" },
   { 4105, DMEC, AE, "Too many Ndb objects" },
   { 4106, DMEC, AE, "All Not NULL attribute have not been defined" },
   { 4114, DMEC, AE, "Transaction is already completed" },
@@ -685,6 +741,7 @@ ErrorBundle ErrorCodes[] = {
   { 4553, DMEC, AE, "NdbLockHandle original operation not executed successfully" },
   { 4554, DMEC, AE, "NdbBlob can only be closed from Active state" },
   { 4555, DMEC, AE, "NdbBlob cannot be closed with pending operations" },
+  { 4556, DMEC, AE, "RecordSpecification has illegal value in column_flags" },
 
   { 4200, DMEC, AE, "Status Error when defining an operation" },
   { 4201, DMEC, AE, "Variable Arrays not yet supported" },
